from functools import wraps, lru_cache

import numpy as np
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord


class hasCache:
    """ Parent Class for Classes that use the cache property """

    def cache_clear(self):
        for func in dir(self):
            if func.startswith("__"):
                continue
            func = getattr(self, func)
            if hasattr(func, "cache_clear"):
                func.cache_clear()


def cache(function):
    """ Cache Property that uses the lru_cache for numpy arrays. Only works with 1 argument however """
    # Switch order of array and self, because only the first argument is used for the cache
    @lru_cache(maxsize=32)
    def cached_wrapper(hashable_array, self):
        if hashable_array is not None:
            array = np.array(hashable_array)
            return function(self, array)
        else:
            return function(self)

    @wraps(function)
    def wrapper(self, array=None):
        if isinstance(array, np.ndarray):
            if array.ndim > 0:
                return cached_wrapper(tuple(array), self)
            else:
                return cached_wrapper(float(array), self)
        else:
            return cached_wrapper(array, self)

    # copy lru_cache attributes over too
    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


def resets_cache(function):
    """ Property for functions that should reset the cache, because they change the result """

    @wraps(function)
    def wrapper(self, value):
        if self._orbit is not None:
            self._orbit.cache_clear()
        return function(self, value)

    return wrapper


def time_input(function):
    """ Checks that the input value is an astropy Time """

    @wraps(function)
    def wrapper(self, value):
        value = Time(value)
        return function(self, value)

    return wrapper


def coord_input(function):
    """ Converts the input of a function into a SkyCoord object """

    @wraps(function)
    def wrapper(self, value):
        value = SkyCoord(value)
        return function(self, value)

    return wrapper


def fget(name):
    name = f"_{name}"

    def f(self):
        return getattr(self, name)

    return f


def fset(name, unit):
    name = f"_{name}"

    try:
        unit = u.Unit(unit)
        decorator = u.quantity_input(value=unit)
    except TypeError:
        if unit is Time:
            decorator = time_input
        elif unit is SkyCoord:
            decorator = coord_input
        else:
            raise TypeError(f"Expected a Quantity, Time, or SkyCoord, but got {unit}")

    @decorator
    def f(self, value):
        setattr(self, name, value)

    return f


def CollectionFactory(cls):
    """ Decorator that turns Collection _fields into properties """

    # Add properties to the class
    for name, alt_names, unit, _, doc in cls._fields:
        getter = fget(name)
        setter = fset(name, unit)

        setattr(cls, name, property(getter, setter, None, doc))
        for alt in alt_names:
            setattr(cls, alt, property(getter, setter))

    cls._names = [f[0] for f in cls._fields]

    return cls



from astropy import units as u
from astropy import constants as c
from astropy.time import Time

from .bodies import Star, Planet

# Section 1: Solar System bodies
# NOTE: inclination is set to 0 for Earth so that we get transits
Sun = Star(mass=c.M_sun, radius=c.R_sun, name="Sun", teff=5770 * u.K)
Earth = Planet(
    mass=c.M_earth,
    radius=c.R_earth,
    sma=1 * u.AU,
    period=1 * u.year,
    ecc=0.016,
    name="Earth",
)
# Earth on a perfectly circular orbit
Earth_circular = Planet(
    mass=c.M_earth, radius=c.R_earth, sma=1 * u.AU, period=1 * u.year, name="Earth"
)

# Section 2: Exoplanet systems
GJ1214 = Star(
    mass=0.157 * c.M_sun, radius=0.2110 * c.R_sun, name="GJ1214", teff=3030 * u.K
)
GJ1214_b = Planet(
    mass=0.0204 * c.M_jup,
    radius=0.239 * c.R_jup,
    sma=0.01433 * u.AU,
    period=1.58 * u.day,
    time_of_transit=Time(54980.248796, format="mjd"),
    name="GJ1214 b",
)



import pytest
import numpy as np

from astropy.time import Time
from astropy import units as u


@pytest.fixture
def times(planet):
    t0 = planet.time_of_transit.mjd
    p = planet.period.to_value("day")
    t = np.linspace(t0 - p / 2, t0 + p / 2, 1001)
    t = Time(t, format="mjd")
    return t


@pytest.fixture
def t0(planet):
    return planet.time_of_transit


def test_mean_anomaly(orbit, times, t0):
    m = orbit.mean_anomaly(times)
    assert isinstance(m, u.quantity.Quantity)
    assert m.to("rad") is not None
    assert m.ndim == times.ndim
    assert m.size == times.size

    m = orbit.mean_anomaly(t0)
    assert isinstance(m, u.quantity.Quantity)
    assert m.to("rad") is not None
    assert m == 0 * u.rad


def test_true_anomaly(orbit, times, t0):
    f = orbit.true_anomaly(times)
    assert isinstance(f, u.quantity.Quantity)
    assert f.to("rad") is not None
    assert f.ndim == times.ndim
    assert f.size == times.size
    assert np.all((f <= np.pi * u.rad) & (f >= -np.pi * u.rad))

    f = orbit.true_anomaly(t0)
    assert isinstance(f, u.quantity.Quantity)
    assert f.to("rad") is not None
    # assert f == 0 # Close to 0?


def test_eccentric_anomaly(orbit, times, t0):
    ea = orbit.eccentric_anomaly(times)
    assert isinstance(ea, u.quantity.Quantity)
    assert ea.to("rad") is not None
    assert ea.ndim == times.ndim
    assert ea.size == times.size
    assert np.all((ea <= np.pi * u.rad) & (ea >= -np.pi * u.rad))

    ea = orbit.eccentric_anomaly(t0)
    assert isinstance(ea, u.quantity.Quantity)
    assert ea.to("rad") is not None
    assert np.isclose(ea.to_value("rad"), 0)


def test_distance(orbit, times, planet):
    d = orbit.distance(times)
    assert isinstance(d, u.quantity.Quantity)
    assert d.to("m") is not None
    assert d.ndim == times.ndim
    assert d.size == times.size

    min_distance = orbit.periapsis_distance()
    max_distance = orbit.apoapsis_distance()
    assert np.all((d <= max_distance) & (d >= min_distance))


def test_phase_angle(orbit, times, t0):
    pa = orbit.phase_angle(times)
    assert isinstance(pa, u.quantity.Quantity)
    assert pa.to("rad") is not None
    assert pa.ndim == times.ndim
    assert pa.size == times.size
    assert np.all((pa >= -np.pi * u.rad) & (pa <= np.pi * u.rad))

    pa = orbit.phase_angle(t0)
    assert isinstance(pa, u.quantity.Quantity)
    assert pa.to("rad") is not None
    assert pa.to_value("rad") == 0


def test_radius(orbit, times, t0):
    r = orbit.projected_radius(times)
    max_d = orbit.apoapsis_distance()

    assert isinstance(r, u.quantity.Quantity)
    assert r.to(u.m) is not None
    assert r.ndim == times.ndim
    assert r.size == times.size
    assert np.all((r >= 0 * u.m) & (r <= max_d))

    r = orbit.phase_angle(t0)
    b = orbit.impact_parameter()
    assert isinstance(r, u.quantity.Quantity)
    assert r.to("deg") is not None
    assert np.isclose(r.to_value("rad"), b.to_value(1))


def test_position3d(orbit, times):
    # x is towards the observer, z is "north", and y to the "right"
    x, y, z = orbit.position_3D(times)
    d = orbit.distance(times)
    max_d = orbit.apoapsis_distance()

    assert isinstance(x, np.ndarray)
    assert x.ndim == times.ndim
    assert x.size == times.size
    assert np.all(np.abs(x) <= max_d)
    assert isinstance(y, np.ndarray)
    assert y.ndim == times.ndim
    assert y.size == times.size
    assert np.all(np.abs(y) <= max_d)
    assert isinstance(z, np.ndarray)
    assert z.ndim == times.ndim
    assert z.size == times.size
    assert np.all(np.abs(z) <= max_d)

    # Results are self consistent (within numerical uncertainties)
    assert np.allclose(x ** 2 + y ** 2 + z ** 2, d ** 2)


def test_mu(orbit, times):
    mu = orbit.mu(times)

    assert isinstance(mu, np.ndarray)
    assert mu.ndim == times.ndim
    assert mu.size == times.size
    assert np.all(mu <= 1)


def test_contact(orbit):
    t0 = orbit.time_primary_transit()
    t1 = orbit.first_contact()
    t2 = orbit.second_contact()
    t3 = orbit.third_contact()
    t4 = orbit.fourth_contact()

    assert isinstance(t0, Time)
    assert isinstance(t1, Time)
    assert isinstance(t2, Time)
    assert isinstance(t3, Time)
    assert isinstance(t4, Time)
    # Assure that the order is correct
    assert t1 < t2 < t0 < t3 < t4


def test_impact_parameter(orbit, star):
    b = orbit.impact_parameter()
    r_s = star.radius
    assert isinstance(b, u.quantity.Quantity)
    assert b.to(1) is not None
    assert 0 <= b <= 1



"""
Changes:
  * Custom field_property
  * Custom dataclass function to automatically annotate properties.
  * Modified _get_field to get default value for property.
  * No Changes to _process_class, included to use _get_field and _init_fn
"""
from copy import copy
import sys
import types
import inspect
from typing import Callable, Any
from dataclasses import (
    Field,
    _FIELD,
    MISSING,
    field,
    fields,
    asdict,
    InitVar,
    _is_classvar,
    _FIELD_CLASSVAR,
    _is_type,
    _is_initvar,
    _FIELD_INITVAR,
    _PARAMS,
    _DataclassParams,
    _FIELDS,
    _POST_INIT_NAME,
    _set_new_attribute,
    _repr_fn,
    _tuple_str,
    _cmp_fn,
    _frozen_get_del_attr,
    _hash_action,
    _init_fn,
)

import numpy as np
from astropy.units.quantity import Quantity
from astropy import units

__all__ = [
    "field_property",
    "get_return_type",
    "dataclass",
    "_get_field",
    "_process_class",
]

# We just need to add a field to property so we can set a default value
class _property(property):
    default_attr = MISSING
    default_factory_attr = MISSING


def get_return_type(
    default: Any = MISSING, default_factory: Callable[[], Any] = MISSING
):
    """Find the return type for the default value or default_factory.

    Args:
        default (object)[MISSING]: is the default value of the field.
        default_factory (callable/function)[MISSING]: is a 0-argument function called to initialize a field's value.

    Returns:
        return_type (type/object)[MISSING]: The type of the default or MISSING if not found.
    """
    # Fields must have an annotation
    return_type = inspect.signature(default_factory).return_annotation
    if isinstance(return_type, units.UnitBase):
        return_type = Quantity
    if return_type == inspect.Signature.empty:
        return_type = MISSING
        if default != MISSING:
            return_type = type(default)
        elif default_factory != MISSING:
            try:
                return_type = type(default_factory())
            except (ValueError, TypeError, Exception):
                pass
    return return_type


class field_property(property):

    get_return_type = staticmethod(get_return_type)

    def __init__(
        self,
        fget: Callable[[Any], Any] = None,
        fset: Callable[[Any, Any], None] = None,
        fdel: Callable[[Any], None] = None,
        doc: str = None,
        default: Any = MISSING,
        default_factory: Callable[[], Any] = MISSING,
    ):

        self.default_attr = default
        self.default_factory_attr = default_factory
        self.name = None
        super().__init__(fget, fset, fdel, doc=doc)

    def __set_name__(self, owner, name):
        try:
            if self.name is None:
                self.name = name
        except (AttributeError, Exception):
            pass

    def getter(self, fget: Callable[[Any], Any]) -> "field_property":
        return type(self)(
            fget,
            self.fset,
            self.fdel,
            self.__doc__,
            default=self.default_attr,
            default_factory=self.default_factory_attr,
        )

    def setter(self, fset: Callable[[Any, Any], None]) -> "field_property":
        return type(self)(
            self.fget,
            fset,
            self.fdel,
            self.__doc__,
            default=self.default_attr,
            default_factory=self.default_factory_attr,
        )

    def deleter(self, fdel: Callable[[Any], None]) -> "field_property":
        return type(self)(
            self.fget,
            self.fset,
            fdel,
            self.__doc__,
            default=self.default_attr,
            default_factory=self.default_factory_attr,
        )

    def default(self, default: Any) -> "field_property":
        self.default_attr = default
        return self

    def default_factory(
        self, default_factory: Callable[[Any], None]
    ) -> "field_property":
        self.default_factory_attr = default_factory
        return self

    __call__ = getter


def create_fset(name, unit):

    if isinstance(unit, units.UnitBase):

        @units.quantity_input
        def fset(self, value: Quantity[unit]):
            setattr(self, name, value)

    else:

        def fset(self, value: unit):
            if value is not None:
                if type(unit) == type and isinstance(value, unit):
                    pass
                else:
                    value = unit(value)
            setattr(self, name, value)

    return fset


def create_fget(name):
    def fget(self) -> Quantity:
        return getattr(self, name)

    return fget


def dataclass(
    cls=None,
    *,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
):
    """Returns the same class as was passed in, with dunder methods
    added based on the fields defined in the class.

    Examines PEP 526 __annotations__ to determine fields.

    If init is true, an __init__() method is added to the class. If
    repr is true, a __repr__() method is added. If order is true, rich
    comparison dunder methods are added. If unsafe_hash is true, a
    __hash__() method function is added. If frozen is true, fields may
    not be assigned to after instance creation.
    """

    def wrap(cls):
        # Annotate all properties
        if not hasattr(cls, "__annotations__"):
            cls.__annotations__ = {}

        # Turn certain values into properties
        for name, attr in cls.__annotations__.items():
            if isinstance(attr, InitVar):
                continue
            # For quantity we add the quantity_inputs decorator
            __name = f"__dataclass_{name}__"
            value = getattr(cls, name)
            # We need to copy it, so its not the same object
            if not isinstance(value, Field):
                value = copy(value)

            if isinstance(attr, units.UnitBase):
                converter = attr
                attr = Quantity
            elif (
                hasattr(attr, "__origin__")
                and issubclass(attr.__origin__, Quantity)
                and hasattr(attr, "__metadata__")
                and len(attr.__metadata__) == 1
                and isinstance(attr.__metadata__[0], units.UnitBase)
            ):
                # Its an annotated type
                converter = attr.__metadata__[0]
                attr = attr.__origin__
            elif issubclass(attr, Quantity):
                converter = value.unit
            elif issubclass(attr, np.ndarray):
                converter = np.asarray
            else:
                converter = attr
            fset = create_fset(__name, converter)
            fget = create_fget(__name)

            # The default value of the property is defined via the
            # default_attr value, which is not available in the
            # default class, thus we use a subclass
            prop = _property(fget, fset)
            if hasattr(value, "default_factory"):
                prop.default_factory_attr = value.default_factory
                setattr(cls, __name, None)
            else:
                prop.default_attr = value
                setattr(cls, __name, value)

            setattr(cls, name, prop)
            cls.__annotations__[name] = attr

        # Turn properties into fields for dataclass
        for name, attr in cls.__dict__.items():
            if (
                isinstance(attr, property)
                and name not in cls.__annotations__
                and attr.fset is not None
            ):
                return_type = get_return_type(default_factory=attr.fget)
                if return_type != MISSING:
                    cls.__annotations__[name] = return_type

        return _process_class(cls, init, repr, eq, order, unsafe_hash, frozen)

    # See if we're being called as @dataclass or @dataclass().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    return wrap(cls)


def _get_field(cls, a_name, a_type):
    # Return a Field object for this field name and type.  ClassVars
    # and InitVars are also returned, but marked as such (see
    # f._field_type).

    # If the default value isn't derived from Field, then it's only a
    # normal default value.  Convert it to a Field().
    default = getattr(cls, a_name, MISSING)
    if isinstance(default, Field):
        f = default
    else:
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = MISSING
        elif isinstance(default, property):
            default_factory_attr = getattr(default, "default_factory_attr", MISSING)
            default_attr = getattr(default, "default_attr", MISSING)

            if default_factory_attr is not MISSING:
                if isinstance(default_factory_attr, (staticmethod, classmethod)):
                    default_factory_attr = default_factory_attr.__get__(cls, cls)
                f = field(default_factory=default_factory_attr)
            elif default_attr is not MISSING:
                f = field(default=default_attr)
            else:
                return_type = inspect.signature(default.fget).return_annotation
                if return_type != inspect.Signature.empty:
                    default = return_type
                    f = field(default_factory=default)
                else:
                    default = MISSING
                    f = field(default=default)
        else:
            f = field(default=default)

    # Only at this point do we know the name and the type.  Set them.
    f.name = a_name
    f.type = a_type

    # Assume it's a normal field until proven otherwise.  We're next
    # going to decide if it's a ClassVar or InitVar, everything else
    # is just a normal field.
    f._field_type = _FIELD

    # In addition to checking for actual types here, also check for
    # string annotations.  get_type_hints() won't always work for us
    # (see https://github.com/python/typing/issues/508 for example),
    # plus it's expensive and would require an eval for every stirng
    # annotation.  So, make a best effort to see if this is a ClassVar
    # or InitVar using regex's and checking that the thing referenced
    # is actually of the correct type.

    # For the complete discussion, see https://bugs.python.org/issue33453

    # If typing has not been imported, then it's impossible for any
    # annotation to be a ClassVar.  So, only look for ClassVar if
    # typing has been imported by any module (not necessarily cls's
    # module).
    typing = sys.modules.get("typing")
    if typing:
        if _is_classvar(a_type, typing) or (
            isinstance(f.type, str)
            and _is_type(f.type, cls, typing, typing.ClassVar, _is_classvar)
        ):
            f._field_type = _FIELD_CLASSVAR

    # If the type is InitVar, or if it's a matching string annotation,
    # then it's an InitVar.
    if f._field_type is _FIELD:
        # The module we're checking against is the module we're
        # currently in (dataclasses.py).
        dataclasses = sys.modules[__name__]
        if _is_initvar(a_type, dataclasses) or (
            isinstance(f.type, str)
            and _is_type(f.type, cls, dataclasses, dataclasses.InitVar, _is_initvar)
        ):
            f._field_type = _FIELD_INITVAR

    # Validations for individual fields.  This is delayed until now,
    # instead of in the Field() constructor, since only here do we
    # know the field name, which allows for better error reporting.

    # Special restrictions for ClassVar and InitVar.
    if f._field_type in (_FIELD_CLASSVAR, _FIELD_INITVAR):
        if f.default_factory is not MISSING:
            raise TypeError(f"field {f.name} cannot have a " "default factory")
        # Should I check for other field settings? default_factory
        # seems the most serious to check for.  Maybe add others.  For
        # example, how about init=False (or really,
        # init=<not-the-default-init-value>)?  It makes no sense for
        # ClassVar and InitVar to specify init=<anything>.

    # For real fields, disallow mutable defaults for known types.
    if f._field_type is _FIELD and isinstance(f.default, (list, dict, set)):
        raise ValueError(
            f"mutable default {type(f.default)} for field "
            f"{f.name} is not allowed: use default_factory"
        )

    return f


def _process_class(cls, init, repr, eq, order, unsafe_hash, frozen):
    # Now that dicts retain insertion order, there's no reason to use
    # an ordered dict.  I am leveraging that ordering here, because
    # derived class fields overwrite base class fields, but the order
    # is defined by the base class, which is found first.
    fields = {}

    if cls.__module__ in sys.modules:
        globals = sys.modules[cls.__module__].__dict__
    else:
        # Theoretically this can happen if someone writes
        # a custom string to cls.__module__.  In which case
        # such dataclass won't be fully introspectable
        # (w.r.t. typing.get_type_hints) but will still function
        # correctly.
        globals = {}

    setattr(cls, _PARAMS, _DataclassParams(init, repr, eq, order, unsafe_hash, frozen))

    # Find our base classes in reverse MRO order, and exclude
    # ourselves.  In reversed order so that more derived classes
    # override earlier field definitions in base classes.  As long as
    # we're iterating over them, see if any are frozen.
    any_frozen_base = False
    has_dataclass_bases = False
    for b in cls.__mro__[-1:0:-1]:
        # Only process classes that have been processed by our
        # decorator.  That is, they have a _FIELDS attribute.
        base_fields = getattr(b, _FIELDS, None)
        if base_fields:
            has_dataclass_bases = True
            for f in base_fields.values():
                fields[f.name] = f
            if getattr(b, _PARAMS).frozen:
                any_frozen_base = True

    # Annotations that are defined in this class (not in base
    # classes).  If __annotations__ isn't present, then this class
    # adds no new annotations.  We use this to compute fields that are
    # added by this class.
    #
    # Fields are found from cls_annotations, which is guaranteed to be
    # ordered.  Default values are from class attributes, if a field
    # has a default.  If the default value is a Field(), then it
    # contains additional info beyond (and possibly including) the
    # actual default value.  Pseudo-fields ClassVars and InitVars are
    # included, despite the fact that they're not real fields.  That's
    # dealt with later.
    cls_annotations = cls.__dict__.get("__annotations__", {})

    # Now find fields in our class.  While doing so, validate some
    # things, and set the default values (as class attributes) where
    # we can.
    cls_fields = [_get_field(cls, name, type) for name, type in cls_annotations.items()]
    for f in cls_fields:
        fields[f.name] = f

        # If the class attribute (which is the default value for this
        # field) exists and is of type 'Field', replace it with the
        # real default.  This is so that normal class introspection
        # sees a real default value, not a Field.
        if isinstance(getattr(cls, f.name, None), Field):
            if f.default is MISSING:
                # If there's no default, delete the class attribute.
                # This happens if we specify field(repr=False), for
                # example (that is, we specified a field object, but
                # no default value).  Also if we're using a default
                # factory.  The class attribute should not be set at
                # all in the post-processed class.
                delattr(cls, f.name)
            else:
                setattr(cls, f.name, f.default)

    # Do we have any Field members that don't also have annotations?
    for name, value in cls.__dict__.items():
        if isinstance(value, Field) and not name in cls_annotations:
            raise TypeError(f"{name!r} is a field but has no type annotation")

    # Check rules that apply if we are derived from any dataclasses.
    if has_dataclass_bases:
        # Raise an exception if any of our bases are frozen, but we're not.
        if any_frozen_base and not frozen:
            raise TypeError("cannot inherit non-frozen dataclass from a " "frozen one")

        # Raise an exception if we're frozen, but none of our bases are.
        if not any_frozen_base and frozen:
            raise TypeError("cannot inherit frozen dataclass from a " "non-frozen one")

    # Remember all of the fields on our class (including bases).  This
    # also marks this class as being a dataclass.
    setattr(cls, _FIELDS, fields)

    # Was this class defined with an explicit __hash__?  Note that if
    # __eq__ is defined in this class, then python will automatically
    # set __hash__ to None.  This is a heuristic, as it's possible
    # that such a __hash__ == None was not auto-generated, but it
    # close enough.
    class_hash = cls.__dict__.get("__hash__", MISSING)
    has_explicit_hash = not (
        class_hash is MISSING or (class_hash is None and "__eq__" in cls.__dict__)
    )

    # If we're generating ordering methods, we must be generating the
    # eq methods.
    if order and not eq:
        raise ValueError("eq must be true if order is true")

    if init:
        # Does this class have a post-init function?
        has_post_init = hasattr(cls, _POST_INIT_NAME)

        # Include InitVars and regular fields (so, not ClassVars).
        flds = [f for f in fields.values() if f._field_type in (_FIELD, _FIELD_INITVAR)]
        _set_new_attribute(
            cls,
            "__init__",
            _init_fn(
                flds,
                frozen,
                has_post_init,
                # The name to use for the "self"
                # param in __init__.  Use "self"
                # if possible.
                "__dataclass_self__" if "self" in fields else "self",
                globals,
            ),
        )

    # Get the fields as a list, and include only real fields.  This is
    # used in all of the following methods.
    field_list = [f for f in fields.values() if f._field_type is _FIELD]

    if repr:
        flds = [f for f in field_list if f.repr]
        _set_new_attribute(cls, "__repr__", _repr_fn(flds, globals))

    if eq:
        # Create _eq__ method.  There's no need for a __ne__ method,
        # since python will call __eq__ and negate it.
        flds = [f for f in field_list if f.compare]
        self_tuple = _tuple_str("self", flds)
        other_tuple = _tuple_str("other", flds)
        _set_new_attribute(
            cls,
            "__eq__",
            _cmp_fn("__eq__", "==", self_tuple, other_tuple, globals=globals),
        )

    if order:
        # Create and set the ordering methods.
        flds = [f for f in field_list if f.compare]
        self_tuple = _tuple_str("self", flds)
        other_tuple = _tuple_str("other", flds)
        for name, op in [
            ("__lt__", "<"),
            ("__le__", "<="),
            ("__gt__", ">"),
            ("__ge__", ">="),
        ]:
            if _set_new_attribute(
                cls, name, _cmp_fn(name, op, self_tuple, other_tuple, globals=globals)
            ):
                raise TypeError(
                    f"Cannot overwrite attribute {name} "
                    f"in class {cls.__name__}. Consider using "
                    "functools.total_ordering"
                )

    if frozen:
        for fn in _frozen_get_del_attr(cls, field_list, globals):
            if _set_new_attribute(cls, fn.__name__, fn):
                raise TypeError(
                    f"Cannot overwrite attribute {fn.__name__} "
                    f"in class {cls.__name__}"
                )

    # Decide if/how we're going to create a hash function.
    hash_action = _hash_action[
        bool(unsafe_hash), bool(eq), bool(frozen), has_explicit_hash
    ]
    if hash_action:
        # No need to call _set_new_attribute here, since by the time
        # we're here the overwriting is unconditional.
        cls.__hash__ = hash_action(cls, field_list, globals)

    if not getattr(cls, "__doc__"):
        # Create a class doc-string.
        cls.__doc__ = cls.__name__ + str(inspect.signature(cls)).replace(" -> None", "")

    return cls



"""
This module describes the datastructures for stars and planets
It also includes functions that are not specific for the orbit, 
but rather the individual bodies

Orbit calculations are performed in the orbit module
"""
import astropy.units as u
from astropy.constants import G, R
from astropy.coordinates import SkyCoord
from astropy.io.misc import yaml
from astropy.time import Time
from astropy.units.quantity import Quantity
from numpy import pi

from .dataclasses_quantity import dataclass


@dataclass
class DataclassIO:
    def to_dict(self) -> dict:
        fields = self.__dataclass_fields__
        data = {k: getattr(self, k) for k in fields.keys()}
        return data

    @classmethod
    def from_dict(cls, data: dict) -> "DataclassIO":
        return cls(**data)

    def save(self, filename: str):
        data = self.to_dict()
        with open(filename, "w") as f:
            yaml.dump(data, f)

    @classmethod
    def load(cls, filename: str) -> "DataclassIO":
        with open(filename, "r") as f:
            data = yaml.load(f)
        return cls.from_dict(data)


@dataclass
class Body(DataclassIO):
    #:str: name of this object
    name: str = ""
    #:Quantity(kg): total mass of the object
    mass: Quantity[u.kg] = 0 * u.kg
    #:Quantity(km): radius of the disk
    radius: Quantity[u.km] = 0 * u.km

    @property
    @u.quantity_input
    def area(self) -> Quantity[u.km ** 2]:
        """Quantity(km**2): area of the disk"""
        return pi * self.radius ** 2

    @property
    @u.quantity_input
    def circumference(self) -> Quantity[u.km]:
        """Quantity(km): circumference of the disk"""
        return 2 * pi * self.radius

    @property
    @u.quantity_input
    def gravity_value(self) -> Quantity[u.m ** 3 / u.s ** 2]:
        """Quantity(): specific gravity parameter for this object"""
        return G * self.mass

    @property
    @u.quantity_input
    def volume(self) -> Quantity[u.km ** 3]:
        """Quantity(km**3): volume of this object"""
        return 4 / 3 * pi * self.radius ** 3

    @property
    @u.quantity_input
    def density(self) -> Quantity[u.kg / u.km ** 3]:
        """Quantity(kg/km**3): mass density of the object"""
        return self.mass / self.volume


@dataclass
class Star(Body):
    # TODO: Surface gravity could be represented more accurately as u.LogUnit("cm/s**2")
    # And Metallicity as u.Unit("dex")

    #:Quantity(K): effective temperature of the star
    teff: Quantity[u.K] = 5700 * u.K
    #:Quantity(log(cgs)): surface gravity of the star
    logg: Quantity[u.one] = 4.4 * u.one
    #:Quantity(1): metalicity of the star, in log units relative to the sun
    monh: Quantity[u.one] = 0 * u.one
    #:Quantity(km/s): micro turbulence parameter
    vmic: Quantity[u.km / u.s] = 1 * u.km / u.s
    #:Quantity(km/s): macro turbulence parameter
    vmac: Quantity[u.km / u.s] = 2 * u.km / u.s
    #:Quantity(km/s): projected rotational velocity of the stellar surface
    vsini: Quantity[u.km / u.s] = 2 * u.km / u.s
    #:SkyCoord: coordinates of the star in the sky
    coordinates: SkyCoord = None
    #:Quantity(pc): distance of the star from the solar system
    distance: Quantity[u.pc] = 0 * u.pc
    #:Quantity(km/s): radial velocity of the star relative to the barycentric
    radial_velocity: Quantity[u.km / u.s] = 0 * u.km / u.s


@dataclass
class Planet(Body):
    #:Quantity(AU): semi major axis of the orbit, i.e. average distance
    sma: Quantity[u.AU] = 1 * u.AU
    #:Quantity(day): orbital period, i.e. 1 year of time on the planet
    period: Quantity[u.day] = 365.25 * u.day
    #:Quantity(1): eccentricity of the orbit, 0 for circular, 1 for parabolic
    ecc: Quantity[u.one] = 0 * u.one
    #:Quantity(deg): inclination of the orbit, where 90deg is transiting
    inc: Quantity[u.deg] = 90 * u.deg
    #:Quantity(deg): argument of periastron, where 90 deg is facing the observer
    omega: Quantity[u.deg] = 90 * u.deg
    #:Time: one known time of transit
    time_of_transit: Time = Time(0, format="mjd")
    #:Quantity(day): duration of the transit, from first contact to fourth contact
    transit_duration: Quantity[u.day] = 0 * u.day

    @property
    def t0(self) -> Time:
        return self.time_of_transit

    @property
    @u.quantity_input
    def surface_gravity(self) -> Quantity[u.m / u.s ** 2]:
        """Quantity(): surface gravity in normal (non logarithmic) units"""
        return self.gravity_value / self.radius ** 2

    @property
    def planet_class(self) -> str:
        """str: archetype of the planet, e.g. gas-giant, rocky"""
        if self.mass > 10 * u.M_earth:
            return "gas-giant"
        else:
            return "rocky"

    @property
    @u.quantity_input
    def atm_molar_mass(self) -> (u.g / u.mol):
        """Quantity: average molar mass of the atmosphere"""
        if self.planet_class == "gas-giant":
            # Hydrogen (e.g. for gas giants)
            atm_molar_mass = 2.5 * (u.g / u.mol)
        elif self.planet_class == "rocky":
            # dry air (mostly nitrogen)  (e.g. for terrestial planets)
            atm_molar_mass = 29 * (u.g / u.mol)
        else:
            raise ValueError
        return atm_molar_mass

    @property
    @u.quantity_input
    def atm_surface_pressure(self) -> u.bar:
        """Quantity: atmopshere pressure at the surface"""
        if self.planet_class == "gas-giant":
            p0 = 0.01 * u.bar
        elif self.planet_class == "rocky":
            p0 = 1 * u.bar
        else:
            raise ValueError
        return p0

    @u.quantity_input
    def equilibrium_temperature(self, stellar_teff: u.K, stellar_radius: u.km) -> u.K:
        """
        Calculate the equilibrium temperature of the planet based on the 
        stellar radiation. Assuming no reflection, i.e. Albedo = 0.

        Parameters
        ----------
        stellar_teff : u.K
            effective temperature of the star

        Returns
        -------
        equi_temp: u.K
            equilibrium temperature of the planet
        """
        t_eq = stellar_teff * (stellar_radius / (2 * self.sma))**0.5
        return t_eq

    @u.quantity_input
    def atm_scale_height(self, surface_temp: u.K) -> u.km:
        """
        Calculate the scale height of the atmosphere, based on the surface
        temperature of the planet. The surface temperature can for example
        be calculated using the equilibrium temperature.

        Parameters
        ----------
        surface_temp : u.K
            surface temperature of the planet

        Returns
        -------
        H: u.km
            scale height of the atmosphere
        """
        teff = surface_temp
        atm_scale_height = (
            R * teff * self.radius ** 2 / (G * self.mass * self.atm_molar_mass)
        )
        return atm_scale_height


if __name__ == "__main__":
    b = Body()
    s = Star()
    p = Planet(name="WASP107b")

    p.save("test.yaml")
    p2 = Planet.load("test.yaml")
    pass



from typing import Tuple
from exoorbit.bodies import Planet, Star
import numpy as np
from scipy.optimize import minimize
from scipy.constants import G, c
from astropy import constants as const
from astropy import units as u
from astropy.time import Time

from .util import time_input

pi = np.pi

# based on http://sredfield.web.wesleyan.edu/jesse_tarnas_thesis_final.pdf
# and the exoplanet handbook


# TODO: Perturbations by other planets
# TODO: relativistic effects?
# TODO: Determine the speed of the cache, is it too slow?
# TODO: Empty cache when self.parameters change


class Orbit:
    def __init__(self, star: Star, planet: Planet):
        """Calculates the orbit of an exoplanet

        Parameters
        ----------
        star : Star
            Central body (star) of the system
        planet : Planet
            Orbiting body (planet) of the system
        """
        self.star = star
        self.planet = planet
        self.star._orbit = self
        self.planet._orbit = self

        # TODO define these parameters
        self.v_s = 0
        self.albedo = 1

    @property
    def a(self) -> u.km:
        return self.planet.sma

    @property
    def p(self) -> u.day:
        return self.planet.period

    @property
    def e(self) -> u.one:
        return self.planet.ecc

    @property
    def i(self) -> u.deg:
        return self.planet.inc

    @property
    def w(self) -> u.deg:
        return self.planet.omega

    @property
    def t0(self) -> Time:
        return self.planet.t0

    @property
    def r_s(self) -> u.km:
        return self.star.radius

    @property
    def m_s(self) -> u.kg:
        return self.star.mass

    @property
    def r_p(self) -> u.km:
        return self.planet.radius

    @property
    def m_p(self) -> u.kg:
        return self.planet.mass

    @property
    def k(self) -> u.one:
        return self.r_p / self.r_s

    @time_input
    def z(self, t: Time) -> u.one:
        """projected radius in units of the stellar disk"""
        return self.projected_radius(t) / self.r_s

    @u.quantity_input
    def periapsis_distance(self) -> u.km:
        """Closest distance in the orbit"""
        return (1 - self.e) * self.a

    @u.quantity_input
    def apoapsis_distance(self) -> u.km:
        """Furthest distance in the orbit"""
        return (1 + self.e) * self.a

    @time_input
    def mean_anomaly(self, t: Time) -> u.rad:
        m = 2 * pi * (t - self.t0) / self.p
        m = m * u.rad
        return m

    @time_input
    def true_anomaly(self, t: Time) -> u.rad:
        root = np.sqrt((1 + self.e) / (1 - self.e))
        ea = self.eccentric_anomaly(t)
        f = 2 * np.arctan(root * np.tan(ea / 2))
        return f

    @time_input
    def eccentric_anomaly(self, t: Time) -> u.rad:
        m = self.mean_anomaly(t)

        tolerance = 1e-8 * m.unit
        e = 0 * m.unit
        en = 10 * tolerance
        while np.any(np.abs(en - e) > tolerance):
            e = en
            en = m + self.e * np.sin(e) * m.unit

        en = ((en + np.pi * u.rad) % (2 * np.pi * u.rad)) - np.pi * u.rad
        return en

    @time_input
    def distance(self, t: Time) -> u.km:
        """Distance from the center of the star to the center of the planet

        Parameters
        ----------
        t : float, array
            time in mjd

        Returns
        -------
        distance : float, array
            distance in km
        """
        return self.a * (1 - self.e * np.cos(self.eccentric_anomaly(t)))

    @time_input
    def phase_angle(self, t: Time) -> u.rad:
        """
        The phase angle describes the angle between
        the vector of observer’s line-of-sight and the
        vector from star to planet

        Parameters
        ----------
        t : float, array
            observation times in jd

        Returns
        -------
        phase_angle : float
            phase angle in radians
        """
        # Determine whether the time is before or after transit
        p = self.p.to_value(u.day)
        k = (t - self.t0).jd % p
        k = k / p
        k = np.where(k < 0.5, 1, -1)
        # Calculate the angle
        f = self.true_anomaly(t)
        theta = np.arccos(np.sin(self.w + f) * np.sin(self.i))
        theta *= k
        return theta

    @time_input
    def projected_radius(self, t: Time) -> u.km:
        """
        Distance from the center of the star to the center of the planet,
        i.e. distance projected on the stellar disk

        Parameters
        ----------
        t : float, array
            time in mjd

        Returns
        -------
        r : float, array
            distance in km
        """
        theta = self.phase_angle(t)
        d = self.distance(t)
        r = np.abs(d * np.sin(theta))
        return r

    @time_input
    def position_3D(self, t: Time) -> Tuple[u.Quantity, u.Quantity, u.Quantity]:
        """Calculate the 3D position of the planet

        the coordinate system is centered in the star, x is towards the observer, z is "north", and y to the "right"

          z ^
            |
            | -¤-
            |̣_____>
            /      y
           / x

        Parameters:
        ----------
        t : float, array
            time in mjd

        Returns
        -------
        x, y, z: float, array
            position in stellar radii
        """
        # TODO this is missing the argument of periapsis
        phase = self.phase_angle(t)
        r = self.distance(t)
        i = self.i
        x = -r * np.cos(phase) * np.sin(i)
        y = -r * np.sin(phase)
        z = -r * np.cos(phase) * np.cos(i)
        return x, y, z

    @time_input
    def mu(self, t: Time) -> u.one:
        # mu = np.cos(self.phase_angle(t))
        r = self.projected_radius(t) / self.r_s
        r = r.decompose()
        mu = np.full_like(r, -1.0)
        mu[r <= 1] = np.sqrt(1 - r[r <= 1] ** 2)
        mu[r > 1] = -np.sqrt(-1 + r[r > 1] ** 2)
        return mu

    @time_input
    def stellar_surface_covered_by_planet(self, t: Time) -> u.one:
        d = self.projected_radius(t)
        area = np.zeros(len(t)) << u.one

        # Use these to make it a bit more readable
        R = self.star.radius
        r = self.planet.radius
        # Case 1: planet completely inside the disk
        area[d + r <= R] = r ** 2 / R ** 2
        # Case 2: planet completely outside the disk
        area[d - r >= R] = 0
        # Case 3: inbetween
        select = (d + r > R) & (d - r < R)
        dp = d[select]
        area1 = r ** 2 * np.arccos((dp ** 2 + r ** 2 - R ** 2) / (2 * dp * r))
        area2 = R ** 2 * np.arccos((dp ** 2 + R ** 2 - r ** 2) / (2 * dp * R))
        area3 = (
            0.5
            * np.sqrt((-dp + r + R) * (dp + r - R) * (dp - r + R) * (dp + r + R))
            * u.rad
        )
        area[select] = (area1 + area2 - area3) / (np.pi * u.rad * R ** 2)

        return area

    def _find_contact(self, r: u.km, bounds: Tuple[float, float]) -> Time:
        func = lambda t: np.abs(
            (self.projected_radius(Time(t, format="mjd")) - r).value
        )
        bounds = [bounds[0].mjd, bounds[1].mjd]
        t0 = bounds[0] + (bounds[1] - bounds[0]) / 4
        res = minimize(func, [t0], bounds=[bounds], method="Powell")
        res = Time(res.x, format="mjd")
        return res

    def first_contact(self) -> Time:
        """
        First contact is when the outer edge of the planet touches the stellar disk,
        i.e. when the transit curve begins

        Returns
        -------
        t1 : float
            time in mjd
        """
        t0 = self.time_primary_transit()
        r = self.r_s + self.r_p
        b = (t0 - self.p / 4, t0)
        return self._find_contact(r, b)

    def second_contact(self) -> Time:
        """
        Second contact is when the planet is completely in the stellar disk for the first time

        Returns
        -------
        t2 : float
            time in mjd
        """
        t0 = self.time_primary_transit()
        r = self.r_s - self.r_p
        b = (t0 - self.p / 4, t0)
        return self._find_contact(r, b)

    def third_contact(self) -> Time:
        """
        Third contact is when the planet begins to leave the stellar disk,
        but is still completely within the disk

        Returns
        -------
        t3 : float
            time in mjd
        """
        t0 = self.time_primary_transit()
        r = self.r_s - self.r_p
        b = (t0, t0 + self.p / 4)
        return self._find_contact(r, b)

    def fourth_contact(self) -> Time:
        """
        Fourth contact is when the planet completely left the stellar disk

        Returns
        -------
        t4 : float
            time in mjd
        """
        t0 = self.time_primary_transit()
        r = self.r_s + self.r_p
        b = (t0, t0 + self.p / 4)
        return self._find_contact(r, b)

    @time_input
    def transit_depth(self, t: Time) -> u.one:
        # r / r_s
        z = self.z(t)
        # r_p / r_s
        k = self.k

        depth = np.zeros(t.shape)

        # Planet fully inside the stellar disk
        depth[z <= (1 - k)] = k ** 2

        # Planet is entering or exiting the disk
        mask = (abs(1 - k) < z) & (z <= (1 + k))
        if np.any(mask):
            z = z[mask]
            k2, z2 = k ** 2, z ** 2
            kappa1 = np.arccos((1 - k2 + z2) / (2 * z))
            kappa0 = np.arccos((k2 + z2 - 1) / (2 * k * z))
            root = 0.5 * np.sqrt(4 * z2 - (1 - k2 + z2) ** 2)
            depth[mask] = 1 / pi * (k2 * kappa0 + kappa1 - root)

        return depth

    @u.quantity_input
    def impact_parameter(self) -> u.one:
        """
        The impact parameter is the shortest projected distance during a transit,
        i.e. how close the planet gets to the center of the star

        This will be 0 if the inclination is 90 deg

        Returns
        -------
        b : float
            distance in km
        """
        d = self.a / self.r_s * np.cos(self.i)
        e = (1 - self.e ** 2) / (1 + self.e * np.sin(self.w))
        return d * e

    @u.quantity_input
    def transit_time_total_circular(self) -> u.day:
        """
        The total time spent in transit for a circular orbit,
        i.e. if eccentricity where 0

        This should be the same as first contact to fourth contact
        There is only an analytical formula for the circular orbit, which is why this exists

        Returns
        -------
        t : float
            time in days
        """
        b = self.impact_parameter()
        alpha = self.r_s / self.a * np.sqrt((1 + self.k) ** 2 - b ** 2) / np.sin(self.i)
        return self.p / pi * np.arcsin(alpha)

    @u.quantity_input
    def transit_time_full_circular(self) -> u.day:
        """
        The total time spent in full transit for a circular orbit,
        i.e. the time during which the planet is completely inside the stellar disk
        if eccentricity where 0

        This should be the same as second contact to third contact
        There is only an analytical formula for the circular orbit, which is why this exists

        Returns
        -------
        t : float
            time in days
        """
        b = self.impact_parameter()
        alpha = self.r_s / self.a * np.sqrt((1 - self.k) ** 2 - b ** 2) / np.sin(self.i)
        return self.p / pi * np.arcsin(alpha)

    def time_primary_transit(self) -> Time:
        """
        The time of the primary transit,
        should be the same as t0

        Returns
        -------
        time : float
            time in mjd
        """
        b = (self.t0 - self.p / 4, self.t0 + self.p / 4)
        return self._find_contact(0, b)

    def time_secondary_eclipse(self) -> Time:
        return self.t0 + self.p / 2 * (1 + 4 * self.e * np.cos(self.w))

    @u.quantity_input
    def impact_parameter_secondary_eclipse(self) -> u.one:
        return (
            self.a
            * np.cos(self.i)
            / self.r_s
            * (1 - self.e ** 2)
            / (1 - self.e * np.sin(self.w))
        )

    @time_input
    def reflected_light_fraction(self, t: Time) -> u.one:
        return (
            self.albedo
            / 2
            * self.r_p ** 2
            / self.distance(t) ** 2
            * (1 + np.cos(self.phase_angle(t)))
        )

    def gravity_darkening_coefficient(self) -> u.one:
        t_s = self.star.teff
        return np.log10(G * self.m_s / self.r_s ** 2) / np.log10(t_s)

    @time_input
    def ellipsoid_variation_flux_fraction(self, t: Time) -> u.one:
        beta = self.gravity_darkening_coefficient()
        return (
            beta
            * self.m_p
            / self.m_s
            * (self.r_s / self.distance(t)) ** 3
            * (np.cos(self.w + self.true_anomaly(t)) * np.cos(self.i)) ** 2
        )

    @time_input
    def doppler_beaming_flux_fraction(self, t: Time) -> u.one:
        rv = self.radial_velocity_star(t)
        return 4 * rv / c

    @time_input
    def radial_velocity_planet(self, t: Time) -> u.km / u.s:
        """
        Radial velocity of the planet in the restframe of the host star
        Positive radial velocity means the planet is moving towards the observer,
        and negative values mean the planet is moving away

        Parameters
        ----------
        t : float, array
            times to evaluate in mjd

        Returns
        -------
        rv : float
            radial velocity in m/s
        """
        K = self.radial_velocity_semiamplitude_planet()
        f = self.true_anomaly(t)
        rv = K * (np.cos(self.w + f) + self.e * np.cos(self.w))
        return rv

    @time_input
    def radial_velocity_star(self, t: Time) -> u.km / u.s:
        """Radial velocity of the star

        Parameters
        ----------
        t : float, array
            times to evaluate in mjd

        Returns
        -------
        rv : float
            radial velocity in m/s
        """
        K = self.radial_velocity_semiamplitude()
        f = self.true_anomaly(t)
        rv = K * (np.cos(self.w + f) + self.e * np.cos(self.w))
        return rv

    @u.quantity_input
    def radial_velocity_semiamplitude(self) -> u.km / u.s:
        """Radial velocity semiamplitude of the star

        Returns
        -------
        K : float
            radial velocity semiamplitude in m/s
        """
        m = self.m_p / u.M_jup * ((self.m_s + self.m_p) / u.M_sun) ** (-2 / 3)
        b = np.sin(self.i) / np.sqrt(1 - self.e ** 2)
        t = (self.p / u.year) ** (-1 / 3)
        return 28.4329 * m * b * t * (u.m / u.s)

    @u.quantity_input
    def radial_velocity_semiamplitude_planet(self) -> u.km / u.s:
        """Radial velocity semiamplitude of the planet

        Returns
        -------
        K : float
            radial velocity semiamplitude in m/s
        """
        m = self.m_s / u.M_jup * ((self.m_s + self.m_p) / u.M_sun) ** (-2 / 3)
        b = np.sin(self.i) / np.sqrt(1 - self.e ** 2)
        t = (self.p / u.year) ** (-1 / 3)
        return 28.4329 * m * b * t * (u.m / u.s)



from setuptools import setup, find_packages
import versioneer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="exoorbit",
    author="Ansgar Wehrhahn",
    author_email="ansgar.wehrhahn@physics.uu.se",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AWehrhahn/ExoOrbits",
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)
